#!/usr/bin/env python

from util import *
import optparse
import sys
import os.path

# First, define some filters used in this process
class Scale(Filter):
  """ Scale the input from 0,0
  """

  def __init__(self, scale):
    self.scale = scale

  def apply(self, command):
    result = command.clone()

    if command.X is not None:
      result.X = command.X * self.scale
    if command.Y is not None:
      result.Y = command.Y * self.scale

    return result

class Clip(Filter):
  """ Clip any commands so they do not escape the desired size
  """

  def __init__(self, size_x, size_y):
    self.size_x = size_x
    self.size_y = size_y

  def apply(self, command):
    result = command.clone()

    if command.X is not None:
      result.X = min(max(command.X, 0.0), self.size_x)
    if command.Y is not None:
      result.Y = min(max(command.Y, 0.0), self.size_y)

    return result

# consume the command line arguments first. Details: https://docs.python.org/2/library/optparse.html
parser = optparse.OptionParser(usage="usage: sandify.py [options] <input.gcode>")
parser.add_option('-l', '--loops', dest='loops', default=10,
        help="Select how many copies of the shape you'd like")
parser.add_option('-x', '--size_x', dest='size_x', default=100,
        help="Select the max size for the X axis.")
parser.add_option('-y', '--size_y', dest='size_y', default=100,
        help="Select the max size for the Y axis.")
parser.add_option('-r', '--rotate', dest='rotate',
        help="Select how many degrees to rotate for each copy of the shape. Positive is CW, negative is CCW. Rotates the shape around the size_x/2, size_y/2")
parser.add_option('-s', '--scale', dest='scale',
        help="Select how many percent to scale the shape for each copy of the shape. Positive will grow, negative will shrink. For example, -10 will scale from 100 to 0 in 10 steps")
(options, input_gcode) = parser.parse_args()

# clean up some of these options
options.loops = int(options.loops)
print 'Configured to run for {} loops'.format(options.loops)
options.size_x = float(options.size_x)
options.size_y = float(options.size_y)
print 'Configured to limit the size between (0,0) and ({},{})'.format(options.size_x, options.size_y)
if options.rotate:
    options.rotate = float(options.rotate)
if options.scale:
    options.scale = float(options.scale)

# check for an input file
if not input_gcode or len(input_gcode) != 1 or not os.path.isfile(input_gcode[0]):
    print('ERROR, you need to specify a gcode file as input')
    sys.exit(1)

print 'Reading in {}'.format(input_gcode[0])

input = loadGCode(input_gcode[0])

# Start with a clean slate of gcode
output = GCode()

print 'Performing transformations'

# for each of these angles
for loop in range(options.loops):
    # rotate the input to that angle
    loop_shape = input.clone()

    if options.rotate:
        loop_shape = loop_shape.clone(Translate(-options.size_x/2, -options.size_y/2),
                                      Rotate(options.rotate * loop),
                                      Translate(options.size_x/2.0, options.size_y/2.0))

    if options.scale:
        loop_shape = loop_shape.clone(Translate(-options.size_x/2, -options.size_y/2),
                                      Scale((1.0 + (0.01 * options.scale * loop))),
                                      Translate(options.size_x/2.0, options.size_y/2.0))

    # Clip
    loop_shape = loop_shape.clone(Clip(options.size_x, options.size_y))

    # copy that rotationm to the output
    output.lines += loop_shape.lines

output_name = 'output.gcode'
print 'Saving output gcode to {}'.format(output_name)
saveGCode(output_name, output)
